
!
! This work (Common Community Physics Package), identified by NOAA, NCAR,
! CU/CIRES, is free of known copyright restrictions and is placed in the
! public domain.
!
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
! IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
! FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
! THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
! IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
! CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
!

!>
!! @brief Auto-generated cap module for the chem_solve scheme
!!
!
module chem_solve_cap

    use, intrinsic :: iso_c_binding,                                   &
                      only: c_f_pointer, c_ptr, c_int32_t
    use            :: ccpp_types,                                      &
                      only: ccpp_t
    use            :: ccpp_fields,                                     &
                      only: ccpp_field_get
    use            :: ccpp_errors,                                     &
                      only: ccpp_error, ccpp_debug
    use            :: chem_solve, &
                      only: chem_solve_init,chem_solve_finalize,chem_solve_run
    ! Other modules required, e.g. type definitions
    
       use machine, only: kind_phys
       use GFS_typedefs, only: GFS_statein_type,  GFS_stateout_type,    &
                               GFS_sfcprop_type,                        &
                               GFS_coupling_type, GFS_control_type,     &
                               GFS_grid_type,     GFS_tbd_type,         &
                               GFS_cldprop_type,  GFS_radtend_type,     &
                               GFS_diag_type,     GFS_interstitial_type,&
                               GFS_init_type
       use micm_type_defs


    implicit none

    private
    public :: chem_solve_init_cap,chem_solve_finalize_cap,chem_solve_run_cap

    contains


    function chem_solve_init_cap(ptr) bind(c) result(ierr)

        integer(c_int32_t)         :: ierr
        type(c_ptr), intent(inout) :: ptr

        type(ccpp_t), pointer           :: cdata
        type(c_ptr)                     :: cptr
        integer, allocatable            :: cdims(:)
        real(kind_phys), pointer :: co(:)
        type(chem_type), pointer     :: chem_obj
        character(len=512), pointer :: errmsg
        integer, pointer :: errflg

        ierr = 0

        call c_f_pointer(ptr, cdata)


        call ccpp_field_get(cdata, 'my_volume_mixing_ratio_co', co, ierr)
        if (ierr /= 0) then
            call ccpp_error('Unable to retrieve my_volume_mixing_ratio_co from CCPP data structure')
            return
        end if

        call ccpp_field_get(cdata, 'chemistry_data', cptr, ierr)
        if (ierr /= 0) then
            call ccpp_error('Unable to retrieve chemistry_data from CCPP data structure')
            return
        end if
        call c_f_pointer(cptr, chem_obj)

        call ccpp_field_get(cdata, 'error_message', errmsg, ierr)
        if (ierr /= 0) then
            call ccpp_error('Unable to retrieve error_message from CCPP data structure')
            return
        end if

        call ccpp_field_get(cdata, 'error_flag', errflg, ierr)
        if (ierr /= 0) then
            call ccpp_error('Unable to retrieve error_flag from CCPP data structure')
            return
        end if

        call chem_solve_init(co=co,chem_obj=chem_obj,errmsg=errmsg,errflg=errflg)
        ierr=errflg

    end function chem_solve_init_cap

    function chem_solve_finalize_cap(ptr) bind(c) result(ierr)

        integer(c_int32_t)         :: ierr
        type(c_ptr), intent(inout) :: ptr

        type(ccpp_t), pointer           :: cdata
        type(c_ptr)                     :: cptr
        integer, allocatable            :: cdims(:)


        ierr = 0

        call c_f_pointer(ptr, cdata)



        call chem_solve_finalize()
        

    end function chem_solve_finalize_cap

    function chem_solve_run_cap(ptr) bind(c) result(ierr)

        integer(c_int32_t)         :: ierr
        type(c_ptr), intent(inout) :: ptr

        type(ccpp_t), pointer           :: cdata
        type(c_ptr)                     :: cptr
        integer, allocatable            :: cdims(:)
        real(kind_phys), pointer :: co(:)
        type(chem_type), pointer     :: chem_obj
        character(len=512), pointer :: errmsg
        integer, pointer :: errflg

        ierr = 0

        call c_f_pointer(ptr, cdata)


        call ccpp_field_get(cdata, 'my_volume_mixing_ratio_co', co, ierr)
        if (ierr /= 0) then
            call ccpp_error('Unable to retrieve my_volume_mixing_ratio_co from CCPP data structure')
            return
        end if

        call ccpp_field_get(cdata, 'chemistry_data', cptr, ierr)
        if (ierr /= 0) then
            call ccpp_error('Unable to retrieve chemistry_data from CCPP data structure')
            return
        end if
        call c_f_pointer(cptr, chem_obj)

        call ccpp_field_get(cdata, 'error_message', errmsg, ierr)
        if (ierr /= 0) then
            call ccpp_error('Unable to retrieve error_message from CCPP data structure')
            return
        end if

        call ccpp_field_get(cdata, 'error_flag', errflg, ierr)
        if (ierr /= 0) then
            call ccpp_error('Unable to retrieve error_flag from CCPP data structure')
            return
        end if

        call chem_solve_run(co=co,chem_obj=chem_obj,errmsg=errmsg,errflg=errflg)
        ierr=errflg

    end function chem_solve_run_cap
end module chem_solve_cap
